package Irpg::Main;
=head1 NAME
Irpg/Main.pm - Entry point of the irpg functionnalities

=head1 SYNOPSIS
    use Irpg::Main;

=head1 DESCRIPTION
    Main functions which updates the players, roll the events,
    parse inputs, ...
=cut
use strict;
use warnings;
use Irpg::Utils qw(:DEFAULT &checksplits &duration &backup &writedb &loaddb);
use Irpg::Irc qw(:DEFAULT :interaction $inbytes);
use Exporter qw(import);
our @EXPORT = qw(&init_hashes &rpcheck &parse &penalize &ha &finduser);
our @EXPORT_OK = qw($pausemode $silentmode $primnick $lastreg);

foreach (qw(Quest Action Event Admin Users)) {
    eval "use Irpg::$_";
}

my %commands; # filling at the very bottom

my $rpreport = 0; # constant for reporting top players
my $lastreg = 0; # holds the time of the last reg. cleared every second.
                 # prevents more than one account being registered / second
my $lasttime_ref;
my %split; # holds nick!user@hosts for clients that have been netsplit
my %onchan; # users on game channel
my %auto_login;

my $pausemode = 0; # pausemode on/off flag
my $silentmode = 0; # silent mode 0/1/2/3, see head of file

my $primnick;
my $opts;
my $rps;
my $prev_online;
my $delayed_funcs;
=head1 FUNCTION init_pkg
    This function sets the references to
    options and players hashes.
=over
=item SCALAR (ref)    - reference to the options hash
=item SCALAR (ref)    - reference to the players hash
=item SCALAR (ref)    - reference to the prev_online hash
=item SCALAR (ref)    - reference to the delayed_funcs array
=cut
sub init_pkg {
    ($opts, $rps, $lasttime_ref, $prev_online, $delayed_funcs) = @_;
    $primnick = $opts->{botnick}; # for regain or register checks
    Irpg::Irc::init_pkg(\$silentmode);
    Irpg::Quest::init_pkg($opts, $rps);
    Irpg::Action::init_pkg($opts, $rps, \$primnick);
    Irpg::Event::init_pkg($opts, $rps);
    Irpg::Admin::init_pkg($opts, $rps);
    Irpg::Users::init_pkg($opts, $rps, \%onchan, \$primnick,
                          \$lastreg, \$silentmode, \$pausemode);
}

sub pause_mode {
    my ($userhost, $usernick, $username, $source, @arg) = @_;
    $pausemode = $pausemode ? 0 : 1;
    Irpg::Irc::privmsg("PAUSE_MODE set to $pausemode.",$usernick,1);
}

sub silent_mode {
    my ($userhost, $usernick, $username, $source, @arg) = @_;
    if (!defined($arg[0]) || $arg[0] < 0 || $arg[0] > 3) {
        Irpg::Irc::privmsg("Try: SILENT <mode>", $usernick,1);
    }
    else {
        $silentmode = $arg[0];
        Irpg::Irc::privmsg("SILENT_MODE set to $silentmode.",$usernick,1);
    }
}

sub reloaddb {
    my ($userhost, $usernick, $username, $source, @arg) = @_;
    if (!$pausemode) {
        Irpg::Irc::privmsg("ERROR: Can only use LOADDB while in PAUSE mode.",
                $usernick, 1);
    }
    else {
        $prev_online = loaddb($rps);
        Irpg::Irc::privmsg("Reread player database file; ".scalar(keys(%$rps)).
                " accounts loaded.",$usernick,1);
    }
}



=head1 FUNCTION finduser
    This function returns the online user associated to a nick.
=over
=item SCALAR (string)    - nick
=cut
sub finduser {
    my $nick = shift;
    my $offline = shift;
    return undef if (!defined($nick));
    for my $user (keys(%$rps)) {
        next unless ($rps->{$user}{online} || $offline);
        if ($rps->{$user}{nick} eq $nick) { return $user; }
    }
    return undef;
}

=head1 FUNCTION ha
    This function returns whether a user has admin access or not.
=over
=item SCALAR (string)    - user name
=cut
sub ha { # return 0/1 if username has access
    my $user = shift;
    if (!defined($user) || !exists($rps->{$user})) {
        debug("Error: Attempted ha() for invalid username \"$user\"");
        return 0;
    }
    return $rps->{$user}{isadmin};
}

=head1 FUNCTION penalize
    This function adds TTL to a user and sends a notice
    according to the type of penality
=over
=item SCALAR (string)    - user name
=item SCALAR (string)    - type of penality
=cut
sub penalize {
    my $username = shift;
    return 0 if !defined($username);
    return 0 if !exists($rps->{$username});
    my $type = shift;
    my $pen = 0;
    my $reason = "existing";
    my $pen_key;
    if ($type eq "quit") {
        $pen = int(20 * ($opts->{rppenstep}**$rps->{$username}{level}));
        $reason = undef;
        $pen_key = 'pen_quit';
    }
    elsif ($type eq "nick") {
        $pen = int(30 * ($opts->{rppenstep}**$rps->{$username}{level}));
        $reason = "nick change";
        $pen_key = 'pen_nick';
    }
    elsif ($type eq "align") {
        $pen = int(30 * ($opts->{rppenstep}**$rps->{$username}{level}));
        $reason = 'alignment change';
        $pen_key = 'pen_align';
    }
    elsif ($type eq "chtitle") {
        $pen = int(30 * ($opts->{rppenstep}**$rps->{$username}{level}));
        $reason = 'title change';
        $pen_key = 'pen_title';
    }
    elsif ($type eq "chclass") {
        $pen = $rps->{$username}{level} <= 60 ?
                int( $opts->{rpbase} * ($opts->{rpstep}**$rps->{$username}{level})) :
                int(($opts->{rpbase} * ($opts->{rpstep}**60))
                    + (86400*($rps->{$username}{level} - 60)));
        $reason = 'training';
        $pen_key = 'pen_class';
    }
    elsif ($type eq "privmsg" || $type eq "notice") {
        #$pen = int(shift(@_) * ($opts->{rppenstep}**$rps->{$username}{level}));
        $pen = -int(shift(@_) * $rps->{$username}{level} / 5);
        $reason = undef;
        $pen_key = 'pen_mesg';
    }
    elsif ($type eq "part") {
        $pen = int(200 * ($opts->{rppenstep}**$rps->{$username}{level}));
        $reason = 'parting';
        $pen_key = 'pen_part';
    }
    elsif ($type eq "kick") {
        $pen = int(250 * ($opts->{rppenstep}**$rps->{$username}{level}));
        $reason = 'being kicked';
        $pen_key = 'pen_kick';
    }
    elsif ($type eq "logout") {
        $pen = int(20 * ($opts->{rppenstep}**$rps->{$username}{level}));
        $reason = 'LOGOUT command';
        $pen_key = 'pen_logout';;
    }
    if ($opts->{limitpen} && $pen > $opts->{limitpen}) {
        $pen = $opts->{limitpen};
    }
    $rps->{$username}{$pen_key} += $pen if (exists($rps->{$username}{$pen_key}));
    $rps->{$username}{next} += $pen;
    if (defined($reason)) {
        Irpg::Irc::notice("Penalty of ".duration($pen)." added to your timer for ".
               "$reason.",$rps->{$username}{nick});
    }
    questpencheck($username) if ($pen > 0);
    return 1; # successfully penalized a user! woohoo!
}

sub moveplayers {
    return unless $$lasttime_ref > 1;
    my $onlinecount = grep { $rps->{$_}{online} } keys %$rps;
    return unless $onlinecount;
    for (my $i=0;$i<$opts->{self_clock};++$i) { # this is to simulate 1 move/s
        # temporary hash to hold player positions, detect collisions
        my @questers = movequesters();
        my %positions = ();
        my @who_fought = ();
        for my $player (keys(%$rps)) {
            next unless ($rps->{$player}{online});
            next if (@questers && grep { $player eq $_ } @questers);
            $rps->{$player}{x} += int(rand(3))-1;
            $rps->{$player}{y} += int(rand(3))-1;
            # if player goes over edge, wrap them back around
            if ($rps->{$player}{x} > $opts->{mapx}) { $rps->{$player}{x} = 0; }
            if ($rps->{$player}{y} > $opts->{mapy}) { $rps->{$player}{y} = 0; }
            if ($rps->{$player}{x} < 0) { $rps->{$player}{x} = $opts->{mapx}; }
            if ($rps->{$player}{y} < 0) { $rps->{$player}{y} = $opts->{mapy}; }
            if (exists($positions{$rps->{$player}{x}}{$rps->{$player}{y}}) &&
                !$positions{$rps->{$player}{x}}{$rps->{$player}{y}}{battled}) {
                if ($rps->{$positions{$rps->{$player}{x}}{$rps->{$player}{y}}{user}}{isadmin} &&
                    !$rps->{$player}{isadmin} && rand(100) < 1) {
                    Irpg::Irc::chanmsg("$player encounters ".
                        $positions{$rps->{$player}{x}}{$rps->{$player}{y}}{user}.
                        " and bows humbly.");
                }
                if (rand($onlinecount) < 1) {
                    $positions{$rps->{$player}{x}}{$rps->{$player}{y}}{battled}=1;
                    collision_action($player,
                        $positions{$rps->{$player}{x}}{$rps->{$player}{y}}{user});
                    push @who_fought, $player;
                }
            }
            else {
                $positions{$rps->{$player}{x}}{$rps->{$player}{y}}{battled}=0;
                $positions{$rps->{$player}{x}}{$rps->{$player}{y}}{user}=$player;
            }
            # at the last move, evil users who has not fought
            # chase near players
            if ($i == $opts->{self_clock}-1
                && $rps->{$player}{alignment} eq 'Evil'
                && !(grep { $player eq $_ } @who_fought)) {
                    evilchase($player, \%positions);
            }
        }
    }
}

=head1 FUNCTION rpcheck
    This function checks levels and update database
=cut
sub rpcheck { # check levels, update database
    # check splits hash to see if any split users have expired
    checksplits() if $opts->{detectsplits};
    # send out $freemessages lines of text from the outgoing message queue
    Irpg::Irc::fq();
    # clear registration limiting
    $lastreg = 0;
    my $online = scalar(grep { $rps->{$_}{online} } keys(%$rps));
    # there's really nothing to do here if there are no online users
    return unless $online;

    ### MOVING PLAYERS ###
    moveplayers();
    
    ### ALL MODULES CHECKS ###
    # statements using $rpreport do not bother with scaling by the clock because
    # $rpreport is adjusted by the number of seconds since last rpcheck()
    foreach (qw(Quest Action Event)) {
        eval 'Irpg::'.$_.'::rpcheck($rpreport, $online)';
    }

    ### TOP PLAYERS REPORT ###
    if ($rpreport && $rpreport%36000==0) { # 10 hours
        my @u = sort { $rps->{$b}{level} <=> $rps->{$a}{level} ||
                       $rps->{$a}{next}  <=> $rps->{$b}{next} } keys(%$rps);
        Irpg::Irc::chanmsg("Idle RPG Top Players:") if @u;
        for my $i (0..2) {
            $#u >= $i and
            Irpg::Irc::chanmsg("$u[$i], the level $rps->{$u[$i]}{level} ".
                    "$rps->{$u[$i]}{alignment} $rps->{$u[$i]}{title} ".
                    "$rps->{$u[$i]}{class}->{NAME}, is #".
                    ($i + 1) . "! Next level in ".
                    (duration($rps->{$u[$i]}{next})).".");
        }
        backup($opts);
    }

    ### RETRIEVE NICK (if in use by other user) ###
    if ($rpreport%1800==0) { # 30 mins
        if ($opts->{botnick} ne $primnick) {
            Irpg::Irc::sts($opts->{botghostcmd}) if $opts->{botghostcmd};
            Irpg::Irc::sts("NICK $primnick");
        }
    }

    ### WARN IF PAUSE MODE ###
    if ($rpreport%600==0 && $pausemode) { # warn every 10m
        Irpg::Irc::chanmsg("WARNING: Cannot write database in PAUSE mode!");
    }

    ### TIME RELATED STUFFS ###
    # (time related data, and here is the levelling)

    # $lasttime = time() on successful join to $opts->{botchan}, initial value is 1).
    # do nothing if not connected
    next if ($$lasttime_ref == 1);
    my $curtime=time();

    ## Delayed Funcs ##
    Irpg::Utils::exec_stack_delayed($curtime);

    ## Update Players Data ##
    for my $k (keys(%$rps)) {
        if ($rps->{$k}{online} && exists $rps->{$k}{nick} &&
            $rps->{$k}{nick} && exists $onchan{$rps->{$k}{nick}}) {
            $rps->{$k}{next} -= ($curtime - $$lasttime_ref);
            @{$rps->{$k}{next_a}} = grep { $_ > 0 }
                                    map { $_ - ($curtime - $$lasttime_ref) }
                                    @{$rps->{$k}{next_a}};
            $rps->{$k}{idled} += ($curtime - $$lasttime_ref);
            if ($rps->{$k}{next} < 1) {
                $rps->{$k}{level}++;
                if (!($rps->{$k}{level} % 5)){
                    $rps->{$k}{points}++;
                    Irpg::Irc::notice(
                        "Your hard training paid off, and you have ".
                        "one more point to invest.",
                        $rps->{$k}{nick});
                }
                if ($rps->{$k}{level} > 60) {
                    $rps->{$k}{next} = int(($opts->{rpbase} *
                                         ($opts->{rpstep}**60)) +
                                         (86400*($rps->{$k}{level} - 60)));
                }
                else {
                    $rps->{$k}{next} = int($opts->{rpbase} *
                                         ($opts->{rpstep}**$rps->{$k}{level}));
                }
                Irpg::Irc::chanmsg("$k, the $rps->{$k}{alignment} ".
                        "$rps->{$k}{title} $rps->{$k}{class}->{NAME}, ".
                        "has attained level $rps->{$k}{level}! ".
                        "Next level in ".duration($rps->{$k}{next}).".");
                find_item($k);
                challenge_opp($k);
            }
            if ($rps->{$k}{actions} + @{$rps->{$k}{next_a}} < int($rps->{$k}{level}/10)) {
                Irpg::Action::new_action($k);
            }
        }
        # attempt to make sure this is an actual user, and not just an
        # artifact of a bad PEVAL
    }
    # do not write in pause mode (would log everyone out if the bot failed to connect)
    # if fails to open $opts->{dbfile}, will not update $lasttime and
    # so should have correct values on next rpcheck(). 
    if (!$pausemode && $rpreport%60==0) { writedb($rps); }
    $rpreport += $opts->{self_clock};
    $$lasttime_ref = $curtime;
}


=head1 FUNCTION parse
    This function parses the input from the irc server:
    the event and the users messages,
    and reacts accordingly
=over
=item SCALAR (string)    - input
=cut
sub parse {
    my($in) = shift;
    $Irpg::Irc::inbytes += length($in); # increase parsed byte count
    $in =~ s/[\r\n]//g; # strip all \r and \n
    debug("<- $in");
    my @arg = split(/\s/,$in); # split into "words"
    my $usernick = substr((split(/!/,$arg[0]))[0],1);
    # logged in char name of nickname, or undef if nickname is not online
    my $username = finduser($usernick);
    if (lc($arg[0]) eq 'ping') { Irpg::Irc::sts("PONG $arg[1]",1); }
    elsif (lc($arg[0]) eq 'error') {
        # uh oh, we've been disconnected from the server, possibly before we've
        # logged in the users in %auto_login. so, we'll set those users' online
        # flags to 1, rewrite db, and attempt to reconnect (if that's wanted of
        # us)
        $rps->{$_}{online}=1 for keys(%auto_login);
        writedb($rps);
        return;
    }
    $arg[1] = lc($arg[1]); # original case no longer matters
    if ($arg[1] eq '433' && $opts->{botnick} eq $arg[3]) {
        $opts->{botnick} .= 0;
        Irpg::Irc::sts("NICK $opts->{botnick}");
    }
    elsif ($arg[1] eq 'join') {
        return if (! $opts->{botchan} eq substr($arg[2], 1));
        # %onchan holds time user joined channel. used just to now who is there
        $onchan{$usernick}=time();

        if (grep { $rps->{$_}{nick} eq $usernick } keys %$rps) {
            Irpg::Irc::sts("WHOIS $usernick");
        }

        if ($opts->{'detectsplits'} && exists($split{substr($arg[0],1)})) {
            delete($split{substr($arg[0],1)});
        }
        elsif ($opts->{botnick} eq $usernick) {
            Irpg::Irc::sts("WHO $opts->{botchan}");
            (my $opcmd = $opts->{botopcmd}) =~ s/%botnick%/$opts->{botnick}/eg;
            Irpg::Irc::sts($opcmd);
            $$lasttime_ref = time(); # start rpcheck()
        }
    }
    elsif ($arg[1] eq 'quit') {
        # if we see our nick come open, grab it (skipping queue)
        if ($usernick eq $primnick) { Irpg::Irc::sts("NICK $primnick",1); }
        elsif ($opts->{'detectsplits'} &&
               "@arg[2..$#arg]" =~ /^:\S+\.\S+ \S+\.\S+$/) {
            if (defined($username)) { # user was online
                $split{substr($arg[0],1)}{time}=time();
                $split{substr($arg[0],1)}{account}=$username;
            }
        }
        elsif (defined($username)) {
            $rps->{$username}{online}=0;
            penalize($username,"quit");
        }
        delete($onchan{$usernick});
    }
    elsif ($arg[1] eq 'nick') {
        # if someone (nickserv) changes our nick for us, update $opts->{botnick}
        if ($usernick eq $opts->{botnick}) {
            $opts->{botnick} = substr($arg[2],1);
        }
        # if we see our nick come open, grab it (skipping queue), unless it was
        # us who just lost it
        elsif ($usernick eq $primnick) { Irpg::Irc::sts("NICK $primnick",1); }
        else {
            if (defined($username)) {
                $rps->{$username}{nick} = substr($arg[2],1);
                substr($rps->{$username}{userhost},0,length($rps->{$username}{nick})) =
                    substr($arg[2],1);
                penalize($username,"nick");
            }
            if (exists($onchan{$usernick})) {
                $onchan{substr($arg[2],1)} = delete($onchan{$usernick});
            }
        }
    }
    elsif ($arg[1] eq 'part') {
        if (defined($username)) {
            $rps->{$username}{online}=0;
            penalize($username,"part");
        }
        delete($onchan{$usernick}) if ($opts->{botchan} eq substr($arg[2], 1));
    }
    elsif ($arg[1] eq 'kick') {
        $usernick = $arg[3];
        $username = finduser($usernick);
        $rps->{$username}{online}=0;
        penalize($username,"kick");
        delete($onchan{$usernick}) if ($opts->{botchan} eq $arg[2]);
    }
    # don't penalize /notices to the bot
    elsif ($arg[1] eq 'notice' && $arg[2] ne $opts->{botnick}) {
        penalize($username,"notice",length("@arg[3..$#arg]")-1);
    }
    elsif ($arg[1] eq '001') {
        # send our identify command, set our usermode, join channel
        Irpg::Irc::sts($opts->{botident});
        Irpg::Irc::sts("MODE $opts->{botnick} :$opts->{botmodes}");
        Irpg::Irc::sts("JOIN $opts->{botchan}");
        for (@{$opts->{listenchan}}) {
            Irpg::Irc::sts("JOIN $_");
        }
        $opts->{botchan} =~ s/ .*//; # strip channel key if present
    }
    elsif ($arg[1] eq '315') {
        # 315 is /WHO end. report who we automagically signed online if it will
        # print < 1k of text
        if (keys(%auto_login)) {
            # not a true measure of size, but easy
            if (length("%auto_login") < 1024 && $opts->{senduserlist}) {
                Irpg::Irc::chanmsg(scalar(keys(%auto_login))." users matching ".
                        scalar(keys(%$prev_online))." hosts automatically ".
                        "logged in; accounts: ".join(", ",keys(%auto_login)));
            }
            else {
                Irpg::Irc::chanmsg(scalar(keys(%auto_login))." users matching ".
                        scalar(keys(%$prev_online))." hosts automatically ".
                        "logged in.");
            }
            if ($opts->{voiceonlogin}) {
                my @vnicks = map { $rps->{$_}{nick} } keys(%auto_login);
                while (@vnicks) {
                    Irpg::Irc::sts("MODE $opts->{botchan} +".
                        ('v' x $opts->{modesperline})." ".
                        join(" ",@vnicks[0..(@vnicks % $opts->{modesperline})-1]));
                    splice(@vnicks,0,$opts->{modesperline});
                }
            }
        }
        else { Irpg::Irc::chanmsg("0 users qualified for auto login."); }
        undef($prev_online);
        undef(%auto_login);
    }
    elsif ($arg[1] eq '005') {
        if ("@arg" =~ /MODES=(\d+)/) { $opts->{modesperline}=$1; }
    }
    elsif ($arg[1] eq '352') {
        # 352 is one line of /WHO. check that the nick!user@host exists as a key
        # in %$prev_online, the list generated in loaddb(). the value is the user
        # to login
        $onchan{$arg[7]}=time() if ($opts->{botchan} eq $arg[3]);
        if (exists($prev_online->{$arg[7]."!".$arg[4]."\@".$arg[5]})) {
            $rps->{$prev_online->{$arg[7]."!".$arg[4]."\@".$arg[5]}}{online} = 1;
            $auto_login{$prev_online->{$arg[7]."!".$arg[4]."\@".$arg[5]}}=1;
        }
        elsif (grep { $rps->{$_}{nick} eq $arg[7] } keys %$rps) {
            Irpg::Irc::sts("WHOIS $arg[7]");
        }
    }
    elsif ($arg[1] eq '307') {
        # 307 is the line of /WHOIS which exists if the user is registered
        my @persos = grep { $rps->{$_}{nick} eq $arg[3] } keys %$rps;
        return unless (@persos);
        my $char_name = $persos[0];
        foreach (@persos) {
            $char_name = $_ if $rps->{$_}{lastlogin} > $rps->{$char_name}{lastlogin};
        }
        $rps->{$char_name}{online} = 1;
        Irpg::Irc::chanmsg("Logged in registered user $arg[3] as $char_name.");
    }
    elsif ($arg[1] eq 'privmsg') {
        $arg[0] = substr($arg[0],1); # strip leading : from privmsgs
        $arg[3] = lc(substr($arg[3],1)); # lowercase, strip leading :

        # if the first word of the msg is "botnick:", 
        # we interpret the following as a potential command
        if ($arg[3] eq lc($opts->{botnick}).':') {
            @arg = (@arg[0..2], $opts->{token}.$arg[4], @arg[5..$#arg])
        }
        my $msgtype = '';
        my $source = '';
        my $token_need;
        my $token_given = 0;
        if ($arg[3] =~ m/\Q$opts->{token}\E/) {
            $token_given = 1;
            $arg[3] = substr($arg[3], 1);
        }

        if (lc($arg[2]) eq lc($opts->{botnick})) { # to us, not channel
            $msgtype = 'prv';
            $token_need = 0;
            $source = $usernick;
        }
        else { # in a channel
            $msgtype = 'pub';
            $token_need = 1;
            $source = $arg[2];
        }
        if ((!$token_need || $token_given) &&
            grep { $arg[3] eq $_ } 
                 grep({ $commands{$_}->{$msgtype} } keys(%commands))) {
            # the message is a command valid in $msgtype message
            if (!$commands{$arg[3]}->{adm} || ha($usernick)) {
                # the user has right to execute the command
                if (exists $onchan{$usernick}) {
                    # the user is on the game chan
                    # we execute the command
                    # $arg[0] is userhost
                    # from $arg[4], it is the message content
                    $commands{$arg[3]}->{ref}($arg[0], $usernick, $username,
                                              $source, @arg[4..$#arg]);
                } else {
                    Irpg::Irc::notice("Sorry, you're not in $opts->{botchan}.",
                            $usernick);
                }
            } else {
                Irpg::Irc::notice("You don't have access to ".uc($arg[3]).".", $usernick);
            }
        } elsif ($msgtype eq 'pub') {
            # the message is not a command
            penalize($username,"privmsg",length("@arg[3..$#arg]"));
        }
    }
}

sub help {
    my ($userhost, $usernick, $username, $source, @arg) = @_;
    my $ctxt = $source =~ m/^#/ ? 'pub' : 'prv';
    my $ret = '';
    my $isadmin = defined($username) ? $rps->{$username}{isadmin} : 0;
    if (!exists($arg[0])) {
        $ret = "Available commands : ".
            join(', ', grep({ $commands{$_}->{$ctxt}
                              && $commands{$_}->{adm} <= $isadmin
                              } keys(%commands))).
            ". Do $opts->{token}help <cmds> for help on a specific command. ".
            "For information on the game, see $opts->{helpurl}. ";
    } 
    elsif (exists($commands{$arg[0]})) {
        $ret = $commands{$arg[0]}->{hlp};
    }
    else {
        $ret = "No such command '$arg[0]'.";
    }

    if ($ctxt eq 'prv' && ha($usernick) && !exists($arg[0])) {
        $ret .= "\nAdmin commands URL is $opts->{admincommurl}";
    }
    Irpg::Irc::privmsg($ret, $source);
}


my ($k, $v);
foreach (qw(Quest Action Event Admin Users)) {
    eval 'while (($k,$v) = each %$Irpg::'.$_.'::commands) {$commands{$k} = $v;}';
}
undef $k;
undef $v;

$commands{pause} = {ref => \&pause_mode, adm => 1, prv => 1, pub => 0,
                    hlp => 'PAUSE : toggle pause mode'};

$commands{silent} = {ref => \&silent_mode, adm => 1, prv => 1, pub => 0,
                     hlp => 'SILENT 0..3 : switch between 4 modes of silence'};

$commands{reloaddb} = {ref => \&reloaddb, adm => 1, prv => 1, pub => 0,
                       hlp => 'RELOADDB : force bot to reload db file, rewriting memory.'.
                              'Only in pause mode.'};

$commands{help} = {ref => \&help, adm => 0, prv => 1, pub => 1,
                   hlp => 'HELP [<cmd>] : get list of commands, '.
                          'or help on specific command.'};

1;
