package Irpg::Utils;
=head1 NAME
Irpg/Utils.pm - Useful functions

=head1 SYNOPSIS
    use Irpg::Utils;

=head1 DESCRIPTION
    Handy functions that may be used in any other module of the Irpg project.

=head2 Exports
=over
=item :text Functions to fill strings
=item :data Functions to manipulate database and config file
=back
=cut

use strict;
use warnings;
use Data::Dumper;
use Irpg::Irc qw(:interaction);
use Exporter qw(import);
our @EXPORT = qw(&clog &debug &pronoun);
our @EXPORT_OK = qw(&set_debug_status &daemonize
                    &ts &mksalt &checksplits
                    &duration &pronoun
                    &readconfig &createdb &loaddb &backup &writedb);
our %EXPORT_TAGS = (text=>[qw(&duration &pronoun &mksalt)],
                    data=>[qw(&readconfig &createdb &loaddb &backup &writedb)]);


foreach (<Irpg/Classes/*.pm>) {
    s/Irpg\/Classes\/(\w+)\.pm/$1/;
    eval "require Irpg::Classes::$_";
}


my $_configfile = '.irpg.conf';

my $delayed_funcs = [];
my $opts;

=head1 FUNCTION mksalt
    This function returns a random salt for passwds
=cut
sub mksalt { # generate a random salt for passwds
    join '',('a'..'z','A'..'Z','0'..'9','/','.')[rand(64), rand(64)];
}


=head1 FUNCTION pronoun
    This function returns a pronoun fitting the given gender.
=over
=item SCALAR (int)     - type of pronoun (1->subject, 2->possessive, 3->object)
=item SCALAR (char)    - gender (default neutral)
=back
=cut
sub pronoun {
    my $type = ['subj', 'poss', 'obj']->[shift()-1];
    return undef unless defined $type;
    my $gender = shift;
    if ($gender eq 'm') { return {'subj'=>'he',
                                  'poss'=>'his',
                                  'obj'=>'him'}->{$type};}
    if ($gender eq 'f') { return {'subj'=>'she',
                                  'poss'=>'her',
                                  'obj'=>'her'}->{$type};}
    return {'subj'=>'they',
            'poss'=>'their',
            'obj'=>'them'}->{$type};
}

=head1 FUNCTION duration
    This function returns a human readable duration.
=over
=item SCALAR (int)    - number of seconds
=back
=cut
sub duration { # return human duration of seconds
    my $s = shift;
    return "NA ($s)" if $s !~ /^\d+$/;
    return sprintf("%d day%s, %02d:%02d:%02d",$s/86400,int($s/86400)==1?"":"s",
                   ($s%86400)/3600,($s%3600)/60,($s%60));
}

=head1 FUNCTION ts
    This function returns the timestamp of right now.
=cut
sub ts { # timestamp
    my @ts = localtime(time());
    return sprintf("[%02d/%02d/%02d %02d:%02d:%02d] ",
                   $ts[4]+1,$ts[3],$ts[5]%100,$ts[2],$ts[1],$ts[0]);
}

=head1 FUNCTION debug
    This function writes a message into the debugfile
    and returns the message.
=over
=item SCALAR (string)    - message
=item SCALAR (ref)        - reference to the options hash
=item SCALAR (bool)        - if the process should die (default false)
=back
=cut
sub debug {
    (my $text = shift) =~ s/[\r\n]//g;
    my $die = shift;
    if ($opts->{debug} || $opts->{verbose}) {
        open(DBG,">>$opts->{debugfile}") or do {
            Irpg::Irc::chanmsg("Error: Cannot open debug file: $!");
            return;
        };
        print DBG ts()."$text\n";
        close(DBG);
    }
    if ($die) { die("$text\n"); }
    return $text;
}

=head1 FUNCTION clog
    This function writes a message into the modsfile
    and returns the message.
=over
=item SCALAR (string)    - message
=item SCALAR (ref)        - reference to the options hash
=back
=cut
sub clog {
    my $mesg = shift;
    open(B,">>$opts->{modsfile}") or do {
        debug("Error: Cannot open $opts->{modsfile}: $!");
        Irpg::Irc::chanmsg("Error: Cannot open $opts->{modsfile}: $!");
        return $mesg;
    };
    print B ts()."$mesg\n";
    close(B);
    return $mesg;
}


=head1 FUNCTION daemonize
    This function daemonizes the process.
=cut
sub daemonize {
    # win32 doesn't daemonize (this way?)
    if ($^O eq "MSWin32") {
        print debug("Nevermind, this is Win32, no I'm not.")."\n";
        return;
    }
    use POSIX 'setsid';
    $SIG{CHLD} = sub { };
    fork() && exit(0); # kill parent
    POSIX::setsid() || debug("POSIX::setsid() failed: $!", 1);
    $SIG{CHLD} = sub { };
    fork() && exit(0); # kill the parent as the process group leader
    $SIG{CHLD} = sub { };
    open(STDIN,'/dev/null') || debug("Cannot read /dev/null: $!",1);
    open(STDOUT,'>/dev/null') || debug("Cannot write to /dev/null: $!",1);
    open(STDERR,'>/dev/null') || debug("Cannot write to /dev/null: $!",1);
    # write our PID to $opts->{pidfile}, or return semi-silently on failure
    open(PIDFILE,">$opts->{pidfile}") || do {
        debug("Error: failed opening pid file: $!");
        return;
    };
    print PIDFILE $$;
    close(PIDFILE);
}


=head1 FUNCTION readconfig
    This function reads the config file
    and returns a hash of the options.
=cut
sub readconfig {
    my %opts;
    if (! -e $_configfile) {
        debug("Error: Cannot find $_configfile. Copy it to this directory, ".
              "please.",1);
    }
    else {
        open(CONF,"<$_configfile") or do {
            debug("Failed to open config file $_configfile: $!",\%opts,1);
        };
        my($line,$key,$val);
        while ($line=<CONF>) {
            next() if $line =~ /^#/; # skip comments
            $line =~ s/[\r\n]//g;
            $line =~ s/^\s+//g;
            next() if !length($line); # skip blank lines
            ($key,$val) = split(/\s+/,$line,2);
            $key = lc($key);
            if (lc($val) eq "on" || lc($val) eq "yes") { $val = 1; }
            elsif (lc($val) eq "off" || lc($val) eq "no") { $val = 0; }
            if ($key eq "die") {
                die("Please edit the file $_configfile to setup your bot's ".
                    "options. Also, read the README file if you haven't ".
                    "yet.\n");
            }
            elsif ($key eq "server") { push(@{$opts{servers}},$val); }
            elsif ($key eq "listenchan") { push(@{$opts{listenchan}},$val); }
            else { $opts{$key} = $val; }
        }
        close(CONF);
        $opts = \%opts;
        return $opts;
    }
}


=head1 FUNCTION backup
    This function copy the current state of the dbfile into backup.
=over
=item SCALAR (ref)    - reference to the options hash
=back
=cut
sub backup {
    my $opts = shift;
    if (! -d ".dbbackup/") { mkdir(".dbbackup",0700); }
    if ($^O ne "MSWin32") {
        system("cp $opts->{dbfile} .dbbackup/$opts->{dbfile}".time());
    }
    else {
        system("copy $opts->{dbfile} .dbbackup\\$opts->{dbfile}".time());
    }
}

=head1 FUNCTION loaddb
    This function loads the content of the database into a hash
    and returns a reference to this hash
=over
=item SCALAR (ref)    - reference to the rps hash
=item SCALAR        - is it the startup loading ?
=back
=cut
sub loaddb { # load the players database
    my $rps = shift;
    return unless ($rps);
    my $startup = shift;
    backup($opts);
    my $l;
    my %prev_online = ();
    if (!open(RPS,$opts->{dbfile}) && -e $opts->{dbfile}) {
        Irpg::Irc::sts("QUIT :loaddb() failed: $!");
    }
    my $classname;
    while ($l=<RPS>) {
        chomp($l);
        next if $l =~ /^#/; # skip comments
        my @i = split("\t",$l);
        print Dumper(@i) if @i != 45;
        if (@i != 45) {
            Irpg::Irc::sts("QUIT: Anomaly in loaddb(); line $. of $opts->{dbfile} has ".
                "wrong fields (".scalar(@i).")");
            debug("Anomaly in loaddb(); line $. of $opts->{dbfile} has wrong ".
                "fields (".scalar(@i).")",1);
        }
        if ($startup) { # if not RELOADDB
            if ($i[8]) { $prev_online{$i[7]}=$i[0]; } # log back in
        }
        ($rps->{$i[0]}{pass},
        $rps->{$i[0]}{isadmin},
        $rps->{$i[0]}{level},
        $rps->{$i[0]}{title},
        $rps->{$i[0]}{next},
        $rps->{$i[0]}{nick},
        $rps->{$i[0]}{userhost},
        $rps->{$i[0]}{online},
        $rps->{$i[0]}{idled},
        $rps->{$i[0]}{x},
        $rps->{$i[0]}{y},
        $rps->{$i[0]}{pen_mesg},
        $rps->{$i[0]}{pen_nick},
        $rps->{$i[0]}{pen_part},
        $rps->{$i[0]}{pen_kick},
        $rps->{$i[0]}{pen_quit},
        $rps->{$i[0]}{pen_quest},
        $rps->{$i[0]}{pen_logout},
        $rps->{$i[0]}{pen_title},
        $rps->{$i[0]}{pen_align},
        $rps->{$i[0]}{pen_class},
        $rps->{$i[0]}{created},
        $rps->{$i[0]}{lastlogin},
        $rps->{$i[0]}{item}{amulet},
        $rps->{$i[0]}{item}{charm},
        $rps->{$i[0]}{item}{helm},
        $rps->{$i[0]}{item}{"pair of boots"},
        $rps->{$i[0]}{item}{"pair of gloves"},
        $rps->{$i[0]}{item}{ring},
        $rps->{$i[0]}{item}{"set of leggings"},
        $rps->{$i[0]}{item}{shield},
        $rps->{$i[0]}{item}{tunic},
        $rps->{$i[0]}{item}{weapon},
        $rps->{$i[0]}{stats}{str},
        $rps->{$i[0]}{stats}{con},
        $rps->{$i[0]}{stats}{wis},
        $rps->{$i[0]}{stats}{int},
        $rps->{$i[0]}{stats}{cha},
        $rps->{$i[0]}{stats}{dex},
        $classname,
        $rps->{$i[0]}{points},
        $rps->{$i[0]}{actions},
        $rps->{$i[0]}{alignment},
        $rps->{$i[0]}{gender}) = (@i[1..7],($startup?0:$i[8]),@i[9..$#i]);

        $classname =~ s/ /_/g;
        $rps->{$i[0]}{class} = eval 'Irpg::Classes::'.$classname.
                                    '->new($rps->{$i[0]}{stats})';
        if (! defined($rps->{$i[0]}{class})) {
            $rps->{$i[0]}{class} = eval 'Irpg::Classes:Farmer->new($rps->{$i[0]}{stats})';
        }
        $rps->{$i[0]}{next_a} = ();
    }
    close(RPS);
    debug("loaddb(): loaded ".scalar(keys(%$rps))." accounts, ".
          scalar(keys(%prev_online))." previously online.");

    return unless (open(DELAYS, $opts->{actionsfile}));
    while (<DELAYS>) {
        chomp;
        next if /^#/; # skip comments
        my ($user, @delays) = split;
        push @{$rps->{$user}{next_a}}, @delays;
    }
    close(DELAYS);

    return \%prev_online;
}


=head1 FUNCTION writedb
    This function write the content of the hash of players into the database
=cut
sub writedb {
    my $rps = shift;
    return unless $rps;
    open(RPS,">$opts->{dbfile}") or do {
        Irpg::Irc::chanmsg("ERROR: Cannot write $opts->{dbfile}: $!");
        return 0;
    };
    my $open_delays = open(DELAYS, ">$opts->{actionsfile}");
    debug("ERROR: Cannot write $opts->{actionsfile}: $!") unless ($open_delays);

    print RPS join("\t","# username",
                        "pass",
                        "is admin",
                        "level",
                        "title",
                        "next ttl",
                        "nick",
                        "userhost",
                        "online",
                        "idled",
                        "x pos",
                        "y pos",
                        "pen_mesg",
                        "pen_nick",
                        "pen_part",
                        "pen_kick",
                        "pen_quit",
                        "pen_quest",
                        "pen_logout",
                        "pen_title",
                        "pen_align",
                        "pen_class",
                        "created",
                        "last login",
                        "amulet",
                        "charm",
                        "helm",
                        "boots",
                        "gloves",
                        "ring",
                        "leggings",
                        "shield",
                        "tunic",
                        "weapon",
                        "str",
                        "con",
                        "wis",
                        "int",
                        "char",
                        "dex",
                        "class",
                        "points",
                        "actions",
                        "alignment",
                        "gender")."\n";
    my $k;
    keys(%$rps); # reset internal pointer
    while ($k=each(%$rps)) {
        if (exists($rps->{$k}{next}) && defined($rps->{$k}{next})) {
            print RPS join("\t",$k,
                                $rps->{$k}{pass},
                                $rps->{$k}{isadmin},
                                $rps->{$k}{level},
                                $rps->{$k}{title},
                                $rps->{$k}{next},
                                $rps->{$k}{nick},
                                $rps->{$k}{userhost},
                                $rps->{$k}{online},
                                $rps->{$k}{idled},
                                $rps->{$k}{x},
                                $rps->{$k}{y},
                                $rps->{$k}{pen_mesg},
                                $rps->{$k}{pen_nick},
                                $rps->{$k}{pen_part},
                                $rps->{$k}{pen_kick},
                                $rps->{$k}{pen_quit},
                                $rps->{$k}{pen_quest},
                                $rps->{$k}{pen_logout},
                                $rps->{$k}{pen_title},
                                $rps->{$k}{pen_align},
                                $rps->{$k}{pen_class},
                                $rps->{$k}{created},
                                $rps->{$k}{lastlogin},
                                $rps->{$k}{item}{amulet},
                                $rps->{$k}{item}{charm},
                                $rps->{$k}{item}{helm},
                                $rps->{$k}{item}{"pair of boots"},
                                $rps->{$k}{item}{"pair of gloves"},
                                $rps->{$k}{item}{ring},
                                $rps->{$k}{item}{"set of leggings"},
                                $rps->{$k}{item}{shield},
                                $rps->{$k}{item}{tunic},
                                $rps->{$k}{item}{weapon},
                                $rps->{$k}{stats}{str},
                                $rps->{$k}{stats}{con},
                                $rps->{$k}{stats}{wis},
                                $rps->{$k}{stats}{int},
                                $rps->{$k}{stats}{cha},
                                $rps->{$k}{stats}{dex},
                                $rps->{$k}{class}->{NAME},
                                $rps->{$k}{points},
                                $rps->{$k}{actions},
                                $rps->{$k}{alignment},
                                $rps->{$k}{gender})."\n";
            print DELAYS "$k ".join(' ', @{$rps->{$k}{next_a}})."\n"
                if (defined $rps->{$k}{next_a} && $open_delays);
        }
    }
    close(RPS);
    close(DELAYS) if ($open_delays);
}

=head1 FUNCTION createdb
    This function create a database from scratch,
    filled by answers of the first admin
    and returns a reference of hash of players (of 1 player, actually)
=over
=item SCALAR (ref)    - reference to the options hash
=back
=cut
sub createdb {
    $|=1;
    my $opts = shift;
    my %rps = ();
    print "$opts->{dbfile} does not appear to exist. I'm guessing this is your ".
          "first time using IRPG. Please give an account name that you would ".
          "like to have admin access [$opts->{owner}]: ";
    chomp(my $uname = <STDIN>);
    $uname =~ s/\s.*//g;
    $uname = length($uname)?$uname:$opts->{owner};
    print "Enter a character title for this account: ";
    chomp(my $utitle = <STDIN>);
    $rps{$uname}{title} = substr($utitle,0,30);
    print "Enter a password for this account: ";
    if ($^O ne "MSWin32") {
        system("stty -echo");
    }
    chomp(my $upass = <STDIN>);
    if ($^O ne "MSWin32") {
        system("stty echo");
    }
    $rps{$uname}{pass} = crypt($upass,mksalt());
    $rps{$uname}{next} = $opts->{rpbase};
    $rps{$uname}{nick} = "";
    $rps{$uname}{userhost} = "";
    $rps{$uname}{level} = 0;
    foreach (qw(str con wis int cha dex)) {
        $rps{$uname}{stats}{$_} = 1;
    }
    $rps{$uname}{class} = Irpg::Classes::Farmer->new($rps{$uname}{stats});
    $rps{$uname}{points} = 1;
    $rps{$uname}{actions} = 0;
    $rps{$uname}{online} = 0;
    $rps{$uname}{idled} = 0;
    $rps{$uname}{created} = time();
    $rps{$uname}{lastlogin} = time();
    $rps{$uname}{x} = int(rand($opts->{mapx}));
    $rps{$uname}{y} = int(rand($opts->{mapy}));
    $rps{$uname}{alignment}='Neutral';
    $rps{$uname}{gender}="n";
    $rps{$uname}{isadmin} = 1;
    for my $item ("ring","amulet","charm","weapon","helm",
                  "tunic","pair of gloves","shield",
                  "set of leggings","pair of boots") {
        $rps{$uname}{item}{$item} = 0;
    }
    for my $pen ("pen_mesg","pen_nick","pen_part",
                 "pen_kick","pen_quit","pen_quest",
                 "pen_logout","pen_logout","pen_title",
                  "pen_align","pen_class") {
        $rps{$uname}{$pen} = 0;
    }
    $rps{$uname}{next_a} = ();
    writedb(\%rps);
    print "OK, wrote you into $opts->{dbfile}.\n";
    return \%rps;
}

=head1 FUNCTION checksplits
    Removes expired split hosts from the hash
=over
=item SCALAR (ref)    - reference to the options hash
=item SCALAR (ref)    - reference to the hash of players
=item SCALAR (ref)    - reference to the splits hash
=back
=cut
sub checksplits { # removed expired split hosts from the hash
    my $rps = shift;
    my $split = shift;
    my $host;
    while ($host = each(%$split)) {
        if (time()-$split->{$host}{time} > $opts->{splitwait}) {
            $rps->{$split->{$host}{account}}{online} = 0;
            delete($split->{$host});
        }
    }
}

=head1 FUNCTION execute_delayed
    add an entry in the list of delayed functions
    ordered by time to be called
=over
=item SCALAR - time when the function will be called
=item SCALAR - function to be called
=item ARRAY  - arguments
=back
=cut
sub execute_delayed {
    my ($delay, $func, @args) = @_;
    return unless (defined($delay) && defined($func));
    my $time = $delay+time();
    push @$delayed_funcs, [$time, $func, @args];
    @$delayed_funcs = sort { $a->[0] <=> $b->[0] } @$delayed_funcs;
    return $time;
}

sub exec_stack_delayed {
    my $curtime = shift or return;
    while (@$delayed_funcs
           && $delayed_funcs->[0][0] <= $curtime) {
        my @func = @{shift @$delayed_funcs};
        $func[1]->(@func[2..$#func]);
    }
}

1;
